package org.mmga.clubs.configuration;

import com.alibaba.fastjson2.JSON;
import com.auth0.jwt.exceptions.JWTVerificationException;
import jakarta.servlet.ServletOutputStream;
import jakarta.servlet.http.HttpServletRequest;
import jakarta.servlet.http.HttpServletResponse;
import lombok.NonNull;
import lombok.extern.slf4j.Slf4j;
import org.mmga.clubs.annotations.AuthorizationRequired;
import org.mmga.clubs.entities.BaseResponse;
import org.mmga.clubs.service.UserService;
import org.mmga.clubs.utils.JwtUtils;
import org.springframework.context.annotation.Configuration;
import org.springframework.web.method.HandlerMethod;
import org.springframework.web.servlet.HandlerInterceptor;
import org.springframework.web.servlet.config.annotation.InterceptorRegistry;
import org.springframework.web.servlet.config.annotation.WebMvcConfigurer;

import java.io.IOException;
import java.nio.charset.StandardCharsets;

@Configuration
@Slf4j
public class AuthorizationConfiguration implements HandlerInterceptor, WebMvcConfigurer{
    private final JwtUtils jwtUtils;
    private final UserService userService;
    public AuthorizationConfiguration(JwtUtils jwtUtils, UserService userService){
        this.jwtUtils = jwtUtils;
        this.userService = userService;
    }
    @Override
    public boolean preHandle(@NonNull HttpServletRequest request,@NonNull HttpServletResponse response,@NonNull Object handler) throws Exception {
        if (handler instanceof HandlerMethod handlerMethod){
            request.setAttribute("user", -1);
            String authorization = request.getHeader("Authorization");
            Integer userId;
            AuthorizationRequired authorizationRequired = handlerMethod.getMethodAnnotation(AuthorizationRequired.class);
            try{
                userId = jwtUtils.verifyToken(authorization);
                request.setAttribute("user", userId);
                if (authorizationRequired != null){
                    int permissionId = authorizationRequired.requiredPermissionId();
                    if (permissionId != -1){
                        if (!userService.userHasPermission(userId, permissionId)) {
                            writeAuthorizationFailedResponse(response, null);
                            return false;
                        }
                    }
                }
            }catch (JWTVerificationException e){
                if (authorizationRequired != null){
                    writeAuthorizationFailedResponse(response, e);
                    return false;
                }
            }
        }
        return true;
    }

    private void writeAuthorizationFailedResponse(HttpServletResponse response, JWTVerificationException e) throws IOException {
        log.debug("用户鉴权时出现错误：", e);
        ServletOutputStream outputStream = response.getOutputStream();
        response.addHeader("Content-Encoding", "UTF-8");
        response.addHeader("Content-Type", "application/json; charset=utf-8");
        BaseResponse<Object> err = BaseResponse.failed(401, "token错误");
        outputStream.write(JSON.toJSONString(err).getBytes(StandardCharsets.UTF_8));
        outputStream.close();
    }
    @Override
    public void addInterceptors(InterceptorRegistry registry) {
        registry.addInterceptor(this);
    }
}
