package org.mmga.clubs.controller;

import com.alibaba.fastjson2.JSON;
import com.alibaba.fastjson2.JSONException;
import com.alibaba.fastjson2.JSONObject;
import io.netty.buffer.ByteBuf;
import io.netty.buffer.Unpooled;
import io.swagger.v3.oas.annotations.tags.Tag;
import jakarta.annotation.Nullable;
import jakarta.websocket.*;
import jakarta.websocket.server.PathParam;
import jakarta.websocket.server.ServerEndpoint;
import lombok.NoArgsConstructor;
import lombok.SneakyThrows;
import lombok.extern.slf4j.Slf4j;
import org.mmga.clubs.entities.chess.Room;
import org.mmga.clubs.entities.chess.packet.*;
import org.mmga.clubs.entities.chess.packet.request.*;
import org.mmga.clubs.entities.user.User;
import org.mmga.clubs.service.UserService;
import org.mmga.clubs.utils.JwtUtils;
import org.mmga.clubs.utils.WebSocketUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.scheduling.annotation.Scheduled;
import org.springframework.stereotype.Component;

import java.lang.reflect.Field;
import java.nio.ByteBuffer;
import java.time.Instant;
import java.time.temporal.ChronoUnit;
import java.util.*;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentLinkedQueue;

@Slf4j
@Tag(name = "五子棋", description = "五子棋相关websocket接口")
@ServerEndpoint("/chess/{token}")
@Component
@NoArgsConstructor
public class ChessController {
    private static final ConcurrentHashMap<String, Session> sessions = new ConcurrentHashMap<>();
    private static final ConcurrentHashMap<String, User> users = new ConcurrentHashMap<>();
    private static final ConcurrentHashMap<UUID, Room> rooms = new ConcurrentHashMap<>();
    private static final ConcurrentHashMap<Session, Boolean> pingPong = new ConcurrentHashMap<>();
    private static JwtUtils jwtUtils;
    private static UserService userService;

    @Autowired
    public void setJwtUtils(JwtUtils jwtUtils) {
        ChessController.jwtUtils = jwtUtils;
    }

    @Autowired
    public void setUserService(UserService userService) {
        ChessController.userService = userService;
    }

    private final Queue<String> messageQueue = new ConcurrentLinkedQueue<>();
    private boolean isLogin = false;

    @SneakyThrows
    @OnOpen
    public void onOpen(Session session, @PathParam("token") String token) {
        isLogin = false;
        Optional<Integer> i = jwtUtils.verifyTokenSafe(token);
        if (i.isEmpty()) {
            WebSocketUtils.sendPacket(new ErrorPacket("token验证失败！"), session);
            log.info("closed by token verifier");
            session.close(new CloseReason(CloseReason.CloseCodes.CANNOT_ACCEPT, "token验证失败！"));
            messageQueue.clear();
            return;
        }
        String sessionId = session.getId();
        User user = userService.getUserById(i.get());
        WebSocketUtils.sendPacket(new UserInfoPacket(user), session);
        String existsUser = null;
        int id = user.getId();
        for (Map.Entry<String, User> entry : users.entrySet()) {
            User value = entry.getValue();
            if (value.getId() == id) {
                existsUser = entry.getKey();
                break;
            }
        }
        users.put(sessionId, user);
        sessions.put(sessionId, session);
        isLogin = true;
        if (existsUser != null) {
            users.remove(existsUser);
            Room roomByPlayer = getRoomByPlayer(existsUser);
            if (roomByPlayer != null) {
                roomByPlayer.replace(existsUser, session);
                WebSocketUtils.sendPacket(new PlayerSideAllocationPacket(roomByPlayer.isPlayerWhite(existsUser)), session);
                WebSocketUtils.sendPacket(roomByPlayer.getRoomInfo(users), session);
            }
            Session remove = sessions.remove(existsUser);
            log.info("session closed by another session");
            remove.close(new CloseReason(CloseReason.CloseCodes.NOT_CONSISTENT, "你的账号已在别处登录"));
        }
        for (String s : messageQueue) {
            this.onMessage(s, session);
        }
    }


    @OnMessage
    public void onPong(PongMessage pongMessage, Session session) {
        ByteBuffer applicationData = pongMessage.getApplicationData();
        ByteBuf byteBuf = Unpooled.copiedBuffer(applicationData);
        long l = System.currentTimeMillis() - byteBuf.readLong();
        log.info("client sent pong response, latency is {}ms", l);
        pingPong.put(session, true);
    }

    @OnMessage
    public void onMessage(String message, Session session) {
        if (!isLogin) {
            messageQueue.add(message);
            return;
        }
        try {
            JSONObject jsonObject = JSON.parseObject(message);
            String packetName = jsonObject.getString("name");
            if (packetName == null) {
                WebSocketUtils.sendPacket(new ErrorPacket("包内缺少name字段！"), session);
            }
            Object response = handleMessage(jsonObject, packetName, session);
            if (response != null) {
                WebSocketUtils.sendPacket(response, packetName, session);
            }
        } catch (JSONException e) {
            WebSocketUtils.sendPacket(new ErrorPacket("JSON格式错误"), session);
        }

    }

    @Nullable
    public Object handleMessage(JSONObject object, String packetName, Session session) {
        if (packetName == null) return new ErrorPacket("错误的请求类型");
        try {
            Class<?> packetType = Class.forName("org.mmga.clubs.entities.chess.packet.request." + packetName + "Request");
            JSONObject payloadJson = object.getJSONObject("payload");
            if (payloadJson == null) {
                Field[] declaredFields = packetType.getDeclaredFields();
                if (declaredFields.length == 0) {
                    payloadJson = new JSONObject();
                }
            }
            if (payloadJson == null) return new ErrorPacket("请求类型参数错误！");
            Object payload;
            try {
                payload = payloadJson.to(packetType);
            } catch (Exception e) {
                return new ErrorPacket("请求payload解析失败！");
            }
            if (payload == null) return null;
            return handlePacket(payload, session);
        } catch (ClassNotFoundException e) {
            return new ErrorPacket("错误的请求类型");
        }
    }

    public Object handlePacket(Object payload, Session session) {
        String sessionId = session.getId();
        if (payload instanceof PlayerJoinRequest joinRequest) {
            Room roomByPlayer = getRoomByPlayer(sessionId);
            if (roomByPlayer != null) return new ErrorPacket("你已经加入过一个房间了");
            Room room = rooms.get(joinRequest.roomId());
            if (room == null) return new ErrorPacket("未知的房间");
            Optional<String> join = room.join(session, users.get(sessionId));
            if (join.isPresent()) return new ErrorPacket(join.get());
            room.broadcast(room.getRoomInfo(users));
        }
        if (payload instanceof RoomListRequest) {
            Collection<Room> values = rooms.values();
            return new RoomListPacket(values);
        }
        if (payload instanceof CreateRoomRequest) {
            Room room = new Room();
            UUID roomId = room.getId();
            rooms.put(roomId, room);
            return new RoomCreatedPacket(roomId);
        }
        if (payload instanceof PlaceChessPieceRequest placeChessPieceRequest) {
            int x = placeChessPieceRequest.x();
            int y = placeChessPieceRequest.y();
            Room roomByPlayer = getRoomByPlayer(sessionId);
            if (roomByPlayer == null) return new ErrorPacket("你还没加入房间呢");
            Optional<String> s = roomByPlayer.downPiece(x, y, sessionId);
            if (s.isPresent()) return new ErrorPacket(s.get());
            roomByPlayer.broadcast(roomByPlayer.getRoomInfo(users));
            Optional<Room.WinInfo> winInfo = roomByPlayer.isWin();
            winInfo.ifPresent(info -> {
                roomByPlayer.broadcast(new HasPlayerWinPacket(info));
                if (info.isWhite()) {
                    WebSocketUtils.sendPacketIfPossible(new PlayerWinPacket(), roomByPlayer.getWhiteSession());
                    WebSocketUtils.sendPacketIfPossible(new PlayerLosePacket(), roomByPlayer.getBlackSession());
                } else {
                    WebSocketUtils.sendPacketIfPossible(new PlayerWinPacket(), roomByPlayer.getBlackSession());
                    WebSocketUtils.sendPacketIfPossible(new PlayerLosePacket(), roomByPlayer.getWhiteSession());
                }
                roomByPlayer.setState(Room.RoomState.FINISHED);
                roomByPlayer.broadcast(roomByPlayer.getRoomInfo(users));
            });
        }
        if (payload instanceof ResetRoomRequest) {
            Room roomByPlayer = getRoomByPlayer(sessionId);
            if (roomByPlayer == null) return new ErrorPacket("你还没加入房间呢");
            roomByPlayer.requestRestart(session);
            roomByPlayer.broadcast(roomByPlayer.getRoomInfo(users));
        }
        return null;
    }

    @Nullable
    private static Room getRoomByPlayer(String id) {
        for (Room value : rooms.values()) {
            if (value.isIn(id)) return value;
        }
        return null;
    }

    public static void staticOnClose(Session session) {
        String sessionId = session.getId();
        Room roomByPlayer = getRoomByPlayer(sessionId);
        if (roomByPlayer != null) {
            roomByPlayer.leave(session);
            roomByPlayer.broadcast(roomByPlayer.getRoomInfo(users));
        }
        log.info("session closed {}", sessionId);
        sessions.remove(sessionId);
        users.remove(sessionId);
        pingPong.remove(session);
    }

    @OnClose
    public void onClose(Session session) {
        staticOnClose(session);
    }


    @SneakyThrows
    @Scheduled(fixedDelay = 10000)
    public static void connectionManager() {
        for (Session value : sessions.values()) {
            if (!pingPong.getOrDefault(value, true)) {
                log.info("closed by connectionManager");
                if (value.isOpen()) {
                    value.close(new CloseReason(CloseReason.CloseCodes.CLOSED_ABNORMALLY, "链接延迟过高"));
                    continue;
                }
                staticOnClose(value);
                continue;
            }
            pingPong.put(value, false);
            log.info("send ping to session {}", value.getId());
            value.getAsyncRemote().sendPing(Unpooled.buffer().writeLong(System.currentTimeMillis()).nioBuffer());
        }
        Instant instant = new Date().toInstant();
        Instant before10Min = instant.plus(-10, ChronoUnit.MINUTES);
        rooms.entrySet().stream()
                .filter(e -> e.getValue().getState().equals(Room.RoomState.CREATED))
                .filter(e -> e.getValue().getLastStateChange().toInstant().isBefore(before10Min))
                .map(Map.Entry::getKey)
                .forEach(rooms::remove);
        rooms.values().forEach(Room::tryClean);
    }
}
