package org.mmga.clubs.service;

import jakarta.servlet.ServletOutputStream;
import jakarta.servlet.http.HttpServletResponse;
import lombok.SneakyThrows;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.codec.binary.Hex;
import org.apache.commons.codec.digest.DigestUtils;
import org.mmga.clubs.dao.UserDao;
import org.mmga.clubs.dao.redis.AvatarDao;
import org.mmga.clubs.entities.BaseResponse;
import org.mmga.clubs.entities.Pager;
import org.mmga.clubs.entities.user.*;
import org.mmga.clubs.entities.user.change.UserAvatarOperationVo;
import org.mmga.clubs.entities.user.change.UserChangeAuthVo;
import org.mmga.clubs.entities.user.change.UserChangePasswordVo;
import org.mmga.clubs.entities.user.change.UserRenameVo;
import org.mmga.clubs.entities.verify.VerifyImgVo;
import org.mmga.clubs.entities.verify.VerifyResponseVo;
import org.mmga.clubs.utils.JwtUtils;
import org.mmga.clubs.utils.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.stereotype.Service;
import org.springframework.util.StreamUtils;
import org.springframework.web.multipart.MultipartFile;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.security.DigestInputStream;
import java.security.MessageDigest;
import java.util.Base64;
import java.util.Date;
import java.util.List;

@Service
@Slf4j
public class UserService {
    private final UserDao userDao;
    private final AuthService authService;
    private final JwtUtils jwtUtils;
    private final VerifyCodeService verifyCodeService;
    private final File tempDir;
    private final AvatarDao avatarDao;
    private final File avatarDir;
    private final ClubService clubService;

    @Autowired
    public UserService(UserDao userDao, AuthService authService, JwtUtils jwtUtils, VerifyCodeService verifyCodeService, File tempDir, AvatarDao avatarDao, @Qualifier("avatarDir") File avatarDir, ClubService clubService) {
        this.userDao = userDao;
        this.authService = authService;
        this.jwtUtils = jwtUtils;
        this.verifyCodeService = verifyCodeService;
        this.tempDir = tempDir;
        this.avatarDao = avatarDao;
        this.avatarDir = avatarDir;
        this.clubService = clubService;
    }

    public BaseResponse<Boolean> login(UserLoginVo user, HttpServletResponse response) {
        /*if (verifyCodeService.invalid(user.key(), user.code())) {
            return BaseResponse.failed(403, "验证码错误");
        }*/
        UserVo userVo = userDao.getUser(user.username(), DigestUtils.sha1Hex(user.password()));
        User u = packageUser(userVo);
        if (response != null && u != null) {
            response.addHeader("Set-Authorization", jwtUtils.createToken(u));
        }
        return u == null ? BaseResponse.failed(404, "无效用户") : BaseResponse.success(true);
    }

    public BaseResponse<Boolean> createUser(UserRegVo user, HttpServletResponse response) {
        if (verifyCodeService.invalid(user.key(), user.code())) {
            return BaseResponse.failed(403, "验证码错误");
        }
        String username = user.username();
        if (userDao.countUser(username) > 0) {
            return BaseResponse.failed(409, "用户已存在");
        }
        UserRegResponseVo responseVo = new UserRegResponseVo(user);
        userDao.addUser(responseVo);
        int newUserId = responseVo.getId();
        User newUser = packageUser(userDao.getUserById(newUserId));
        if (newUser != null && response != null) {
            response.addHeader("Set-Authorization", jwtUtils.createToken(newUser));
        }
        return BaseResponse.success(true);
    }

    public User getUserById(int id) {
        return packageUser(userDao.getUserById(id));
    }

    public User getGuestUser() {
        User user = new User();
        user.setId(-1);
        user.setName("游客");
        user.setAuth(authService.getAuthById(1));
        return user;
    }

    private User packageUser(UserVo vo) {
        if (vo == null) {
            return null;
        }
        User result = new User();
        result.setId(vo.id());
        result.setName(vo.name());
        result.setAvatar(vo.avatar());
        result.setAuth(authService.getAuthById(vo.auth()));
        result.setClub(clubService.getClubByUserId(vo.id()));
        return result;
    }

    public BaseResponse<User> getUserInfo(int userId) {
        if (userId == -1) {
            return BaseResponse.success(getGuestUser());
        }
        return BaseResponse.success(getUserById(userId));
    }

    public boolean userHasPermission(int userId, int permissionId) {
        return authService.authHasPermission(userDao.getUserAuthId(userId), permissionId);
    }

    public BaseResponse<Pager<User>> getAllUserInfo(int num, int page) {
        long totalCount = userDao.queryTotalUserCount();
        List<UserVo> user = userDao.getUser((page - 1) * num, num);
        List<User> list = user.stream().map(this::packageUser).toList();
        Pager<User> userPager = new Pager<>(totalCount, list);
        return BaseResponse.success(userPager);
    }

    public BaseResponse<Boolean> changeUsername(UserRenameVo renameVo, int userId) {
        int changeUserId = renameVo.id();
        if (changeUserId == userId || userHasPermission(userId, 4)) {
            String newName = renameVo.newName();
            if (userDao.countUser(newName) > 0) {
                return BaseResponse.failed(409, "用户名已存在！");
            }
            userDao.changeUsername(changeUserId, newName);
            return BaseResponse.success(true);
        }
        return BaseResponse.failed(401, "权限不足，修改失败！", false);
    }

    public BaseResponse<Boolean> changePassword(UserChangePasswordVo changePasswordVo, int userId) {
        int changeUserId = changePasswordVo.getId();
        String oldPassword = changePasswordVo.getOldPassword();
        String newPassword = changePasswordVo.getNewPassword();
        if (changeUserId == userId) {
            if (userDao.changePassword(changeUserId, oldPassword, newPassword) > 0) {
                return BaseResponse.success(true);
            }
            return BaseResponse.failed(404, "旧密码错误！", false);
        }
        if (userHasPermission(userId, 4)) {
            userDao.changePasswordAdmin(changeUserId, newPassword);
            return BaseResponse.success(true);
        }
        return BaseResponse.failed(401, "权限不足，修改失败！", false);
    }

    public BaseResponse<Boolean> changeAuth(UserChangeAuthVo userChangeAuthVo) {
        userDao.changeAuth(userChangeAuthVo.id(), userChangeAuthVo.authId());
        return BaseResponse.success(true);
    }

    public BaseResponse<Boolean> changeAvatar(String avatarOperationCode, int uid) {
        UserAvatarOperationVo byOperationCode = avatarDao.findByOperationCode(avatarOperationCode);
        if (byOperationCode == null) {
            return BaseResponse.failed(404, "操作码错误！");
        }
        int operationUserId = byOperationCode.userId();
        if (operationUserId != uid && !userHasPermission(uid, 4)) {
            return BaseResponse.failed(403, "权限不足");
        }
        avatarDao.delete(byOperationCode);
        String sha = byOperationCode.avatarSHA();
        userDao.changeAvatar(operationUserId, sha);
        File file = new File(avatarDir, sha);
        File tmpFile = new File(tempDir, avatarOperationCode);
        if (!tmpFile.exists()) {
            return BaseResponse.failed(500, "无法找到已上传的头像信息");
        }
        if (!file.exists()) {
            if (!tmpFile.renameTo(file)) {
                log.error("移动头像文件失败！");
            }
        } else {
            if (!tmpFile.delete()) {
                log.error("无法删除临时头像文件");
            }
        }
        return BaseResponse.success(true);
    }

    public BaseResponse<VerifyResponseVo> getVerifyCode() {
        Base64.Encoder mimeEncoder = Base64.getMimeEncoder();
        VerifyImgVo verifyImgVo = verifyCodeService.generateVerifyCode();
        String s = mimeEncoder.encodeToString(verifyImgVo.img());
        return BaseResponse.success(new VerifyResponseVo(s, verifyImgVo.code()));
    }

    @SneakyThrows
    public BaseResponse<String> changeAvatar(MultipartFile file, int userId) {
        InputStream inputStream = file.getInputStream();
        DigestInputStream sha1 = new DigestInputStream(inputStream, MessageDigest.getInstance("SHA1"));
        String operationCode;
        while (true) {
            operationCode = StringUtils.generatorRandomString(32);
            File tmpAvatarImg = new File(tempDir, operationCode);
            if (tmpAvatarImg.exists()) {
                continue;
            }
            FileOutputStream fileOutputStream = new FileOutputStream(tmpAvatarImg);
            StreamUtils.copy(sha1, fileOutputStream);
            fileOutputStream.close();
            sha1.close();
            break;
        }
        String sha1Result = Hex.encodeHexString(sha1.getMessageDigest().digest());
        UserAvatarOperationVo userAvatarOperationVo = new UserAvatarOperationVo(operationCode, sha1Result, userId, new Date());
        avatarDao.save(userAvatarOperationVo);
        return BaseResponse.success(operationCode);
    }

    @SneakyThrows
    public void getAvatar(HttpServletResponse response, String sha1) {
        File avatar = new File(avatarDir, sha1);
        if (avatar.exists()) {
            FileInputStream fileInputStream = new FileInputStream(avatar);
            response.setContentType("image/png");
            ServletOutputStream outputStream = response.getOutputStream();
            StreamUtils.copy(fileInputStream, outputStream);
            outputStream.close();
            fileInputStream.close();
        }
    }

    public UserSimpleInformation getUserSimpleInformation(int userId) {
        return userDao.getUserSimpleInformation(userId);
    }

    public BaseResponse<User> getSimpleInfo(int userId) {
        if (userId == -1) {
            return BaseResponse.failed(404, "用户不存在！");
        }
        return this.getUserInfo(userId);
    }
}
